/*

  yyylowlevelnet - a low level networking library
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYLOWLEVELNET_HPP
#define __YYYLOWLEVELNET_HPP

#include "binarydefs.hpp"

#include <string>

#include <netinet/in.h>

namespace yyynetwork {

  class hostname {
  public:
    hostname (); // initializes to undefined
    hostname (const string &s);
    const string getname () const; // fatal if not defined.
    const bool isdefined () const;
    void undefine ();
  private:
    string rep;
    bool isdefined_rep;
  };

  class ipv4address {
  public:
    ipv4address (); // initializes to undefined
    ipv4address (const string &s); // format like this: 127.23.1.5, object gets undefined if format error or illegal address
    const string dump () const; // format like this: 127.23.1.5, fatal if not defined.
    const struct in_addr & getraw () const;
    void setraw (const struct in_addr & i);
    const bool isdefined () const;
    void undefine();
  private:
    struct in_addr rep;
    bool isdefined_rep;
  };

  class portnumber {
  public:
    portnumber (); // initializes to undefined.
    portnumber (const uint16 pn);
    const uint16 getnumber () const;
    void setnumber (const uint16 pn);
    const bool isdefined () const;
    void undefine();
  private:
    uint16 rep;
    bool isdefined_rep;
  };

  class tcphandle {
  public:
    tcphandle (); // initializes to undefined;
    ~tcphandle (); // automatically performs cleanup (i.e. closing)
    // you cannot copy a tcphandle.
    tcphandle (const tcphandle &);
    tcphandle & operator= (const tcphandle &);
    const bool isopen () const;
    void close();
    const int getfilehandle () const; // fatal if undefined

    void makesocket(); // must not be open, otherwise fatal error.
    // for all functions below, the tcphandle must be open, i.e. makesocket already called, otherwise fatal error.
    void setreuse();
    void setsolinger();
    void setnonblocking();
    enum connect_result_t { CONNECTED, REFUSED, INPROGRESS, UNREACHABLE };
    void connect(const ipv4address &foreign_ipv4address, const portnumber &foreign_portnumber,
		 connect_result_t &res);
    enum bind_result_t { BINDED, ADDRESSINUSE, ACCESSDENIED };
    void bind(const portnumber &local_portnumber, bind_result_t &res);
    void listen();
    enum accept_result_t { ACCEPTED, WOULDBLOCK };
    // newhandle must not be open, but will eventually be afterwards.
    void accept(tcphandle &newhandle, ipv4address &foreign_ipv4address, portnumber &foreign_portnumber,
		accept_result_t &res);
    const ssize_t read(void *const buf, const size_t count);
    const ssize_t write(void *const buf, const size_t count);
  private:
    int handle_rep;
    bool isopen_rep;
    bool mustshutdown;
  };


};

#endif
