/*

  yyyscheduler - a cooperative scheduler
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "yyyscheduler.hpp"

#include "selector.hpp"

#include <algorithm>

using namespace yyyscheduler;

scheduler *scheduler::gsched = 0;

scheduler::scheduler()
  : isterminated_rep(false), curposactive(false)
{
};

scheduler::~scheduler()
{
  if (gsched==this) gsched=0;
};

void scheduler::run()
{
  selector s;
  while (!isterminated()) {
    s.reset();

    for (curpos=notlist.begin();curpos!=notlist.end();++curpos)
      if ((*curpos)->getwatchtype()==ionotify::READ) s.add_read((*curpos)->getfilehandle());
      else if ((*curpos)->getwatchtype()==ionotify::WRITE) s.add_write((*curpos)->getfilehandle());
      else s.add_exception((*curpos)->getfilehandle());

    curpos2=timlist.begin();
    if (curpos2!=timlist.end()) {
      for (list<abstimer *>::iterator i=timlist.begin();i!=timlist.end();++i)
	if ((*i)->firetime() < (*curpos2)->firetime()) curpos2=i;
      s.settimeout((*curpos2)->firetime());
    };

    s.waitnow();

    curposactive=true;
    for (curpos=notlist.begin();curpos!=notlist.end();)
      if ((*curpos)->getwatchtype()==ionotify::READ) {
	if (s.isready_read((*curpos)->getfilehandle())) (*(curpos++))->fire.emit(**curpos);
	else ++curpos;
      }else if ((*curpos)->getwatchtype()==ionotify::WRITE) {
	if (s.isready_write((*curpos)->getfilehandle())) (*(curpos++))->fire.emit(**curpos);
	/* 2000-05-21 13:39:42 CEST                                          ^^^
	   if you leave out the emit and call directly, it will pass a zero pointer,
	   if you optimize with -O2. maybe it is a compiler bug, but i won't ever omit "emit" again ;)
	   otherwise i have to waste more hours searching this sucking error.  (erikyyy) */
	else ++curpos;
      }else{
	if (s.isready_exception((*curpos)->getfilehandle())) (*(curpos++))->fire.emit(**curpos);
	else ++curpos;
      };

    if (s.istimeoutexpired() && (curpos2!=timlist.end())) {
      abstimer &mytimer (**curpos2);
      mytimer.setactive(false);
      mytimer.fire.emit(mytimer);
    };
    curposactive=false;
  };
  setterminated(false);
};

void scheduler::setterminated(const bool isterminated_ = true)
{
  isterminated_rep=isterminated_;
};

void scheduler::register_ionotify (ionotify *const x)
{
  notlist.push_front(x);
};

void scheduler::unregister_ionotify (ionotify *const x)
{
  list<ionotify *>::iterator i=find(notlist.begin(),notlist.end(),x);
  if (curposactive&&(curpos==i)) curpos++;
  notlist.erase(i);
};

void scheduler::register_abstimer (abstimer *const x)
{
  timlist.push_front(x);
};

void scheduler::unregister_abstimer (abstimer *const x)
{
  list<abstimer *>::iterator i=find(timlist.begin(),timlist.end(),x);
  if (curposactive&&(curpos2==i)) curpos2=timlist.end();
  timlist.erase(i);
};


scheduler::genericnotify::genericnotify (scheduler *const sched_)
  : sched (sched_)
{
};


scheduler::ionotify::ionotify (scheduler *const sched_)
  : genericnotify(sched_), isactive_rep(false)
{
};

void scheduler::ionotify::setfilehandle(const int filehandle_, const watchtype_t watchtype_)
{
  setactive(false);
  filehandle_rep=filehandle_;
  watchtype_rep=watchtype_;
  setactive(true);
};

scheduler::ionotify::~ionotify ()
{
  setactive(false);
};

void scheduler::ionotify::setactive(const bool isactive_)
{
  if (isactive()==isactive_) return;
  isactive_rep=isactive_;

  if (isactive()) sched->register_ionotify(this);
  else sched->unregister_ionotify(this);
};


scheduler::abstimer::abstimer (scheduler *const sched_)
  : genericnotify(sched_), isactive_rep(false)
{
};

scheduler::abstimer::~abstimer()
{
  setactive(false);
};

void scheduler::abstimer::setfiretime (const yyyzeit::zeit &firetime_)
{
  setactive(false);
  firetime_rep=firetime_;
  setactive(true);
};

void scheduler::abstimer::setactive (const bool isactive_)
{
  if (isactive()==isactive_) return;
  isactive_rep=isactive_;

  if (isactive()) sched->register_abstimer(this);
  else sched->unregister_abstimer(this);
};
