/*

  yyyqtscheduler - implementation of the coopscheduler concept using qt
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYQTSCHEDULER_HPP
#define __YYYQTSCHEDULER_HPP

#include <sigc++/signal_system.h>

#include "yyyzeit.hpp"

namespace yyyqtscheduler {

class scheduler {
public:
  static scheduler *gsched;

  class genericnotify { // do not use this class. it is internal only.
  public:
    genericnotify(scheduler *const sched_);
    // you can not copy notification objects.
    genericnotify (const genericnotify &);
    genericnotify & operator= (const genericnotify &);
  protected:
    scheduler *const sched;
  };
  
  class ionotify : public genericnotify {
  public:
    ionotify (scheduler *const sched_ = gsched);
    ~ionotify ();

    enum watchtype_t {READ,WRITE,EXCEPTION};
    
    void setfilehandle(const int filehandle_, const watchtype_t watchtype_);
    inline const int getfilehandle() const { return filehandle_rep; };
    inline const watchtype_t getwatchtype() const { return watchtype_rep; };
    
    void setactive(const bool isactive_);
    inline const bool isactive() const { return isactive_rep; };

    SigC::Signal1<void, ionotify &> fire;

  private:
    int filehandle_rep;
    watchtype_t watchtype_rep;
    bool isactive_rep;
  };
  
  class abstimer : public genericnotify {
  public:
    abstimer (scheduler *const sched_ = gsched);
    ~abstimer ();
    
    // setting the firetime also does setactive(true)
    // setactive(false) will be called right before firing.
    void setfiretime (const yyyzeit::zeit &firetime_);
    inline const yyyzeit::zeit &firetime() const { return firetime_rep; };
    
    void setactive(const bool isactive_);
    inline const bool isactive() const { return isactive_rep; };

    SigC::Signal1<void, abstimer &> fire;
    
  private:
    yyyzeit::zeit firetime_rep;
    bool isactive_rep;
  };

  scheduler();
  // you cannot copy a scheduler.
  scheduler (const scheduler &);
  scheduler & operator= (const scheduler &);
  ~scheduler();
};

};

#endif
