#include "memchk.hpp"
#include "yyyexceptions.hpp"
#include "fatal.hpp"
#include "yyygtkscheduler.hpp"
#include "yyypostman.hpp"
#include "yyynetwork.hpp"
#include "yyysendables.hpp"

#include <unistd.h>
#include <stdio.h>
#include <iostream>
#include <gtk--/button.h>
#include <gtk--/main.h>
#include <gtk--/window.h>
#include <gtk--/label.h>
#include <gtk--/box.h>

using SigC::slot;
using namespace yyygtkscheduler;
using namespace yyypostman;
using namespace yyynetwork;
using namespace yyysmartptr;
using namespace yyybinbuff;

void destroy_handler()
{
  Gtk::Main::quit();
}

struct dummy {const bool operator< (const dummy &) const {return false;};};

void appendtobinbuffer (const dummy a, yyybinbuff::binbuff &b)
{
};

const yyybinbuff::binbuff::index readfrombinbuffer(dummy &a, const yyybinbuff::binbuff &b, const yyybinbuff::binbuff::index pos)
{
  return 0;
};

class HelloWorld : public Gtk::Window
{
public:
  Gtk::Button *m_button;
  Gtk::Label *m_label;
  Gtk::Box *vbox;

  typedef connection<scheduler,binbuff> con_t;
  typedef smartptr<con_t> sptr_con_t;
  typedef server<scheduler, binbuff> server_t;
  typedef postman<dummy,con_t> postman_t;
  typedef smartptr<postman_t> sptr_postman_t;

  server_t myserver;
  sptr_con_t mycon;
  sptr_postman_t post;

  HelloWorld()
  {
    myserver.resultavailable.connect(SigC::slot(*this,&HelloWorld::newconnection));
    myserver.setport(10000);

    destroy.connect(slot(&destroy_handler));
    set_border_width(10);
    set_title("SERVER on port 10000");

    vbox = manage(new Gtk::VBox(FALSE,5));
    add(*vbox);
    m_button = manage(new Gtk::Button("Send a big S"));
    m_button->clicked.connect(slot(this, &HelloWorld::hello));
    vbox->pack_start(*m_button,false,false);
    m_label=manage(new Gtk::Label(">>> "));
    vbox->pack_start(*m_label,false,false);

    show_all();
  }

  ~HelloWorld()
  {
    post.kill(); // <-- important that postman gets killed BEFORE the connection object
  };
      
  void hello()
  {
    if (post.isnull()) {
      cout << "i have no connection! please connect me on port 10000. for example with telnet localhost 10000 or with gtkclient.\n";
      return;
    };
    cout << "Sending S...\n";
    post->sender1<char>(dummy())('S');
  }
  
  void newconnection (server_t &, sptr_con_t &c)
  {
    if (mycon.isnull()) {
      mycon=c;
      mycon->sig_disconnected.connect(SigC::slot(*this,&HelloWorld::disconnected));
      post=new postman_t(&(*mycon));
      post->corruptreceived.connect(SigC::slot(*this,&HelloWorld::corruptreceived));
      post->sender1_open<char> (dummy());
      post->receiver1_open<char>(dummy()).connect(SigC::slot(*this,&HelloWorld::recvmsg));
    }else{
      warning("i already have a connection, discarding the new one.");
    };
  };

  void disconnected(con_t &)
  {
    cout << "connection disconnected\n";
    post.kill(); // always first delete the postman, because the postman has a pointer to mycon!
    mycon.kill();
  };

  void corruptreceived(postman_t &)
  {
    cout << "postman corruptreceived\n";
    mycon->disconnect();
  };

  void recvmsg(char m)
  {
    m_label->set_text(m_label->get_text()+m);
    cout << "recvmsg" << m << " !\n";
  };
};





int main(int argc, char *argv[])
{
  memchk::resetsituation(); // gtkmm disables my memchk. email is on way. let's see if they change it.
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      
      Gtk::Main kit(argc, argv);
      HelloWorld helloworld;
      kit.run();
      cout << "yes normal end\n";
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
