/*

  yyyexceptions - some tools for common exceptions
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "yyyexceptions.hpp"

#include "fatal.hpp"

#include <set>
#include <stdio.h>
#include <errno.h>

using namespace yyyexceptions;

static set<const except_base*> li; // there can only be ONE exception in here, but several copies of it!

static void dbg()
{
  set<const except_base*>::iterator i;
  for (i=li.begin();i!=li.end();++i)
    cout << *i << " ";
  cout << "\n";
};

static void myterminatehandler ()
{
  if (li.empty()) {
    fatal ("myterminatehandler() called");
  }else{
    fatal (string("the exception that overwrote the last one is:\n")+(*li.begin())->humandump());
  };
};

void yyyexceptions::except_init ()
{
  set_terminate(myterminatehandler);
};

except_base::except_base ()
{
  cout << "exceptbase cons   ";  dbg();
  if ( ! li.empty())
    warning (string("this exception was overwritten and is now lost!\n") + (*li.begin())->humandump());
  li.clear();
  li.insert(this);
  cout << "basecons done   ";  dbg();
};

except_base::except_base (const except_base &x)
{
  cout << "except_base copy cons   "; dbg();
  if (li.count(&x)) li.insert(this);
  cout << "except_base copy cons end   "; dbg();
};

except_base::~except_base ()
{
  cout << "exceptbase dest    "; dbg();
  li.erase(this);
  cout << "exceptbase dest end   "; dbg();
};


except_normal::except_normal (const string &errormessage_)
  : errormessage(errormessage_)
{
};

string except_normal::humandump() const
{
  return errormessage+string("\n");
};

static const string convint(const int i)
{
  char o[sizeof(i)*4+2];
  sprintf(o,"%d",i);
  return string(o);  
};

static const string mkerrno(const bool disperrno)
{
  if (disperrno) return string("errno = ")+string(strerror(errno))+string("\n"); 
  else return string();
};


except_codepos_internal::except_codepos_internal (const string &errormessage, const string &filename,
						  const int linenumber, const bool disperrno)
  : except_normal (string("Filename: ")+filename+string("  Linenumber: ")+convint(linenumber)+"\n"+mkerrno(disperrno)+errormessage)
{
};
