include(macros.m4)dnl
dnl
dnl specify how many template parameters you want at maximum.
dnl if you want to learn how all this works, set NUMP to 0 or 1 and do m4. then read yyysigdeferer.hpp
define(NUMP,4)dnl
dnl
dnl by the way, you can set your editor to C++ mode for this file.
dnl at least emacs does a very good job.
dnl
/*

  yyysigdeferer - a signal time deferer for libsigc++ and the scheduler concept
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYSIGDEFERER_HPP
#define __YYYSIGDEFERER_HPP

#include "yyyzeit.hpp"
#include "yyytypemagic.hpp"

#include <sigc++/signal_system.h>
#include <map>

/*

  since the objects are sent much later than they are received,
  they need to be copied.
  this happens when the objects are received by the deferer.
  all types are copied as is (more than once), except
  if you use references, the OBJECT itself will be copied ONCE.
  if this is not what you wanted, use pointers, not references.

  example: if you use "int", the int will be copied more than once.
  if you use int& the int& will be copied more than once, but the int
  itself will also be copied, but only ONCE.

  you should always try to use references on copy expensive objects, because if you don't,
  the objects will be copied quite often due to the signal slot implementation
  and other internals.

  in general you need not think, just use.
  the only exception to normal signals and slots or normal function calling
  is that even though you use references, the object will still be copied ONCE.


  all currently buffered signals with the same delay time/type are guaranteed
  to be delivered in the same order than sent.

*/




/*

NOTES FOR FUTURE DEVELOPMENT:

modes of operation:
 absolute point in time
 relative time delay
 don't fire
 fire at once
 fire at once in scheduler

other functions:
 flush (but in which order ???)

*/

namespace yyysigdeferer {

define([<MK_SIGDEFERER>],[<dnl
template <REP($1,class m?-?t,KOMMA )[<>]IFNZ($1,KOMMA )class coopscheduler>
class sigdeferer$1 : public SigC::Object {
public:
  sigdeferer$1 (coopscheduler *const mysched_ = coopscheduler::gsched)
    : mysched(mysched_), delaytype(FIREATONCEINSCHEDULER) {};
  void setreltime(const yyyzeit::zeit &delay_) { delay = delay_; delaytype = RELDELAY; };
  void setabstime(const yyyzeit::zeit &delay_) { delay = delay_; delaytype = ABSDELAY; };
  void setfireatonceinscheduler () { delaytype = FIREATONCEINSCHEDULER; };

  void receive(REP($1,m?-?t m?-?,KOMMA ))
  {
    smart_elem ne ( new elem (REP($1,m?-?,KOMMA )[<>]IFNZ($1,KOMMA )mysched) );
    l[&ne->t]=ne;
    ne->t.fire.connect(slot(*this,&sigdeferer$1::fireslot));
    switch (delaytype) {
    case ABSDELAY: ne->t.setfiretime(delay); break;
    case RELDELAY: ne->t.setfiretime(yyyzeit::getcurrenttime() + delay); break;
    case FIREATONCEINSCHEDULER: ne->t.setfiretime(yyyzeit::zerotime()); break;
    };
  };

  SigC::Signal$1 <void[<>]IFNZ($1,KOMMA )REP($1,m?-?t,KOMMA )> send;

private:
  void fireslot(typename coopscheduler::abstimer &a) {
    maptype::iterator i = l.find(&a);
    send.emit (REP($1,i->second->m?-?,KOMMA ));
    l.erase(i);
  };

  struct elem {
    elem(REP($1,yyytypemagic::add_reference<m?-?t>::type m?-?_,KOMMA
         )[<>]IFNZ($1,KOMMA
         )coopscheduler *const cs)
      : t(cs)[<>]IFNZ($1,KOMMA )REP($1,m?-?(m?-?_),KOMMA ) {};
    typename coopscheduler::abstimer t;
REP($1,[<dnl
    yyytypemagic::remove_reference<m?-?t>::type m?-?;
>])dnl
  };
  typedef yyysmartptr::smartptr <elem> smart_elem;
  typedef map <typename coopscheduler::abstimer *, smart_elem> maptype;
  maptype l;

  coopscheduler *const mysched;
  yyyzeit::zeit delay;

  enum {ABSDELAY, RELDELAY, FIREATONCEINSCHEDULER} delaytype;
};

>])dnl
REP0(NUMP,[<MK_SIGDEFERER(?-?)>])dnl
};

#endif
