#include "memchk.hpp"
#include "yyypostman.hpp"
#include "yyynetwork.hpp"
#include "fatal.hpp"
#include "yyyscheduler.hpp"
#include "yyyexceptions.hpp"
#include "yyysendables.hpp"

#include <vector>

using namespace yyyscheduler;
using namespace yyypostman;
using namespace yyynetwork;
using namespace yyysmartptr;
using namespace yyybinbuff;

class clientprocess : public SigC::Object
{
public:
  typedef connection<scheduler,binbuff> con_t;
  typedef smartptr<con_t> sptr_con_t;
  typedef client<scheduler, binbuff> client_t;
  typedef postman<const uint8,con_t> postman_t;

  sptr_con_t mycon;
  client_t myclient;
  postman_t *post;

  clientprocess ()
    : post(0)
  {
    myclient.netconnect(ipv4address("127.0.0.1"),portnumber(10000)).connect(SigC::slot(*this,&clientprocess::resultavailable));
  };
  ~clientprocess()
  {
    delete post; // <-- important that postman gets killed BEFORE the connection object
  };

  void resultavailable (client_t &, client_t::result &r)
  {
    if (r.isconnected()) {
      mycon=r.getconnection();
      mycon->sig_disconnected.connect(SigC::slot(*this,&clientprocess::disconnected));
      post=new postman_t(&(*mycon));
      post->corruptreceived.connect(SigC::slot(*this,&clientprocess::corruptreceived));
      post->receiver1_open<int> ('x').connect(SigC::slot(*this,&clientprocess::recvint));
      post->receiver1_open<int> ('C').connect(SigC::slot(*this,&clientprocess::recvchar));
      post->receiver1_open<uint8> ('a').connect(SigC::slot(*this,&clientprocess::recv_uint8));
      // not worxing due to libsigc++ post->receiver1_open<const uint8> ('b').connect(SigC::slot(*this,&clientprocess::recv_const_uint8));
      post->receiver1_open<uint8&> ('c').connect(SigC::slot(*this,&clientprocess::recv_uint8_ref));
      post->receiver1_open<const uint8&> ('d').connect(SigC::slot(*this,&clientprocess::recv_const_uint8_ref));
    }else{
      cout << "connection refused\n";
      scheduler::gsched->setterminated(true);
    };
  };

  void recv_uint8 (uint8 a)
  {
    cout << "recv_uint8 " << a << endl;
  };

  void recv_const_uint8 (const uint8 a)
  {
    cout << "recv_const_uint8 " << a << endl;
  };

  void recv_uint8_ref (uint8 &a)
  {
    cout << "recv_uint8_ref " << a << endl;
  };

  void recv_const_uint8_ref (const uint8 &a)
  {
    cout << "recv_const_uint8_ref " << a << endl;
  };

  void disconnected(con_t &)
  {
    cout << "connection disconnected\n";
    scheduler::gsched->setterminated(true);
  };

  void corruptreceived(postman_t &)
  {
    cout << "postman corruptreceived\n";
    mycon->disconnect();
  };

  void recvint(int i)
  {
    cout << "RECVINT " << i << " !!!\n";
  };

  void recvchar(int c)
  {
    cout << char(c) << endl;
  };
};

int main()
{
  memchk::resetsituation();
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      clientprocess cli;
      s.run();
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
