/*

  memchk - memory checking routines
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "memchk.hpp"

#include <stdlib.h>
#include <string.h>

using namespace memchk;

static void fatal (const char *const m1, const char *const m2) throw ()
{
  cerr<<endl<<endl<<"memchk bug detected"<<endl<<m1<<endl<<m2<<endl<<endl<<"terminating."<<endl;
  terminate();
};

/*
struct resblk {
  resblk () : state(UNUSED) {};
  enum { UNUSED=0, ALLOC_NEW, ALLOC_NEWBRACKET, FREED_NEW, FREED_NEWBRACKET } state;
  void *addr;
  size_t size;
};

static const int resanz=10000;
static resblk resbuf[resanz];

void reserve (void *addr, size_t size,bool isbracket)
{
  for (int i=0;i<resanz;++i) {
    if ((resbuf[i].state==resblk::UNUSED)
	|| (resbuf[i].state==resblk::FREED_NEW)
	|| (resbuf[i].state==resblk::FREED_NEWBRACKET)) {
      resbuf[i].state=isbracket ? resblk::ALLOC_NEWBRACKET : resblk::ALLOC_NEW;
      resbuf[i].addr=addr;
      resbuf[i].size=size;
      return;
    };
  };
  fatal("reserve","no space left in resbuf. please increase constant");
};

void unreserve (void *addr, size_t size, bool isbracket)
{
  for (int i=0;i<resanz;++i) {
    switch (resbuf[i].state) {
    case resblk::UNUSED: continue;
    case resblk::ALLOC_NEW:
      if (resbuf[i].addr!=addr) continue;
      if (resbuf[i].size!=size) fatal ("unreserve","shall unreserve memory but sizes don't match");
      if (isbracket) fatal("unreserve","shall unreserve bracket memory but have nonbracket memory");
      resbuf[i].state=resblk::FREED_NEW;
      return;
    case resblk::ALLOC_NEWBRACKET:
      if (resbuf[i].addr!=addr) continue;
      if (resbuf[i].size!=size) fatal ("unreserve","shall unreserve memory but sizes don't match");
      if (!isbracket) fatal("unreserve","shall unreserve nonbracket memory but have bracket memory");
      resbuf[i].state=resblk::FREED_NEWBRACKET;
      return;
    case resblk::FREED_NEW: continue;
    case resblk::FREED_NEWBRACKET: continue;
    };
  };
  fatal("unreserve","the block cannot be found!");
};

void reservelist ()
{
  const char *m;
  for (int i=0;i<resanz;++i) {
    if (resbuf[i].state==resblk::UNUSED) continue;
    switch (resbuf[i].state) {
    case resblk::ALLOC_NEW:        m="ALLOC_NEW        "; break;
    case resblk::ALLOC_NEWBRACKET: m="ALLOC_NEWBRACKET "; break;
    case resblk::FREED_NEW:        m="FREED_NEW        "; break;
    case resblk::FREED_NEWBRACKET: m="FREED_NEWBRACKET "; break;
    case resblk::UNUSED:;
    };
    cout << m << resbuf[i].addr << " " << resbuf[i].size << " B" << endl;
  };
};
*/
#define reserve(a,b,c) ;
#define unreserve(a,b,c) ;
#define reservelist() ;

static char pre_cp[100] = {
  -123,71,31,-105,-119,-61,7,-91,-8,-36,
  50,-56,6,77,64,-115,-61,80,-66,104,
  72,126,43,-81,-62,-6,81,-44,43,118,
  -42,48,61,117,72,71,-71,-49,108,49,
  43,30,121,-79,-20,57,-66,47,9,-4,
  24,-46,-6,-61,1,60,61,-46,-112,-23,
  -55,-26,-103,-122,-36,97,77,21,-80,57,
  -58,92,-40,-65,-115,68,120,-52,-13,1,
  72,-117,83,-61,-50,-43,127,-116,39,-112,
  -11,112,-10,14,119,82,-16,68,-25,32
};
static const size_t pre_cp_size = 100;

static char post_cp[100] = {
  -88,-32,-14,-100,123,64,-109,-42,84,40,
  -57,-123,120,-88,-92,-90,31,108,-71,-5,
  -74,104,121,-40,61,102,95,-121,85,61,
  96,125,-99,-45,-103,-103,-109,-83,-17,104,
  85,54,109,77,94,-111,116,-3,125,-83,
  120,-76,-106,113,12,83,87,-21,90,44,
  -87,58,41,-58,-115,67,-33,-95,112,78,
  -119,69,4,118,19,-30,-120,7,95,-123,
  52,88,-71,74,73,69,29,33,-79,-9,
  -51,-38,-78,119,32,-65,58,-128,-32,42
};
static const size_t post_cp_size = 100;

static const char allocfiller = 0xE2;    // make sure bullshit is in the memory after allocation
static const char deallocfiller = 0x2B;  // make sure bullshit is in the free memory after deletion

struct block {
  size_t size;
  size_t size_not;
  bool allocated;
  bool isbracket;
};

/*
static void stati (bool add, void *blk, size_t sz)
{
  cout << (add ? "alloc " : "free  ");
  cout << blk << " " << sz << " B   ";
  cout << "new ( " << stat_new_mem << " B, " << stat_new_anz << " ) ";
  cout << "new[] ( " << stat_newbracket_mem << " B, " << stat_newbracket_anz << " ) ";
  cout << endl;
};
*/
#define stati(a,b,c) ;

static inline void* allocate (const size_t s, bool isbracket, const char *const who) throw (std::bad_alloc)
{
  char  *buffer    = (char*) malloc ( sizeof(block) + pre_cp_size + s + post_cp_size + sizeof(block) );

  block *preblock  = (block*) (buffer);
  char  *precheck  = ((char*)preblock) + sizeof(block);
  char  *data      = precheck + pre_cp_size;
  char  *postcheck = data + s;
  block *postblock = (block*) (postcheck + post_cp_size);

  preblock->size = s;
  preblock->size_not = ~s;
  preblock->allocated = true;
  preblock->isbracket = isbracket;
  memcpy(precheck,pre_cp,pre_cp_size);

  postblock->size = s;
  postblock->size_not = ~s;
  postblock->allocated = true;
  postblock->isbracket = isbracket;
  memcpy(postcheck,post_cp,post_cp_size);

  memset(data,allocfiller,s);
  
  return data;
};

static inline size_t deallocate (void *p, bool isbracket, const char *const who) throw ()
{
  char *buffer = ((char*)p) - sizeof(block) - pre_cp_size;

  block *preblock  = (block*) (buffer);
  size_t s = preblock->size;
  char  *precheck  = ((char*)preblock) + sizeof(block);
  char  *data      = precheck + pre_cp_size;
  char  *postcheck = data + s;
  block *postblock = (block*) (postcheck + post_cp_size);

  if ( (s != ~preblock->size_not) || (s != postblock->size) || (s != ~postblock->size_not) )
    fatal (who, "deallocate: sizes not consistent");

  if (memcmp(precheck,pre_cp,pre_cp_size))
    fatal (who, "deallocate: data corrupted pre of buffer");

  if (memcmp(postcheck,post_cp,post_cp_size))
    fatal (who, "deallocate: data corrupted post of buffer");
  
  if ( (!preblock->allocated) || (!postblock->allocated) )
    fatal (who, "deallocate: probably deleting already deleted area");
  preblock->allocated = postblock->allocated = false;

  if ( (preblock->isbracket != isbracket) || (postblock->isbracket != isbracket) )
    fatal (who, "deallocate: deleting the wrong kind of memory concerning bracketniess");

  memset(data,deallocfiller,s);
  
  free(buffer);
 
  return s;
};

void *operator new (size_t s) throw (std::bad_alloc)
{
  void *res = allocate (s,false,"void *operator new (size_t) throw (std::bad_alloc)");
  reserve (res,s,false);
  stat_new_mem+=s;
  stat_new_anz++;
  stati(true,res,s);
  return res;
};

void *operator new[] (size_t s) throw (std::bad_alloc)
{
  void *res = allocate (s,true,"void *operator new[] (size_t) throw (std::bad_alloc)");
  reserve (res,s,true);
  stat_newbracket_mem+=s;
  stat_newbracket_anz++;
  stati(true,res,s);
  return res;
};

void operator delete (void *p) throw()
{
  if (!p) return;
  size_t res=deallocate(p,false,"void operator delete (void *) throw()");
  unreserve(p,res,false);
  stat_new_mem-=res;
  stat_new_anz--;
  stati(false,p,res);
};

void operator delete[] (void *p) throw()
{
  if (!p) return;
  size_t res=deallocate(p,true,"void operator delete[] (void *) throw()");
  unreserve(p,res,true);
  stat_newbracket_mem-=res;
  stat_newbracket_anz--;
  stati(false,p,res);
};

void *operator new (size_t, const std::nothrow_t&) throw()
{
  terminate();
};

void *operator new[] (size_t, const std::nothrow_t&) throw()
{
  terminate();
};

void operator delete (void *, const std::nothrow_t&) throw()
{
  terminate();
};

void operator delete[] (void *, const std::nothrow_t&) throw()
{
  terminate();
};

void memchk::resetsituation () throw()
{
  stat_new_mem = stat_newbracket_mem = 0;
  stat_new_anz = stat_newbracket_anz = 0;
};

bool memchk::everythingzero () throw()
{
  return !(stat_new_mem!=0 || stat_newbracket_mem!=0 ||
	   stat_new_anz!=0 || stat_newbracket_anz!=0);
};

void memchk::printstatistics (ostream &os)
{
  os << "memory statistics:"<<endl<<
    "NEW: noch reservierte Bytes: " << stat_new_mem << endl <<
    "  noch reservierte Bloecke: " << stat_new_anz << endl <<
    "NEW[]: noch reservierte Bytes: " << stat_newbracket_mem << endl <<
    "  noch reservierte Bloecke: " << stat_newbracket_anz << endl;
  reservelist();
};

ssize_t memchk::stat_new_mem=0, memchk::stat_newbracket_mem=0;
int memchk::stat_new_anz=0, memchk::stat_newbracket_anz=0;
