#include "yyyscheduler.hpp"
#include "fatal.hpp"
#include "memchk.hpp"
#include "yyynetwork.hpp"
#include "yyysmartptr.hpp"
#include "yyysigdeferer.hpp"
#include "yyyexceptions.hpp"

#include <vector>

#include <unistd.h>

using namespace yyynetwork;
using namespace yyysmartptr;
using namespace yyyscheduler;
using namespace yyysigdeferer;
using namespace yyyzeit;

// define the class "bb". it is a model of the binbuffer concept
class bb {
public:
  typedef int32 index;
  inline const uint8 & operator [] (const index x) const { return rep[x]; };
  inline uint8 & operator [] (const index x) { return rep[x]; };
  inline const index size() const { return rep.size(); };
  inline void resize(const index s) { rep.resize(s); };
  inline void erase (const index first, const index last) { rep.erase(&rep[first], &rep[last]); };
private:
  vector<uint8>rep;
};


class testbed : public SigC::Object {
public:
  typedef connection<scheduler,bb> con_t;
  typedef smartptr<con_t> sptr_con_t;
  typedef client<scheduler, bb> client_t;
  testbed ()
    : cli(scheduler::gsched)
  {
    cli.netconnect(ipv4address("127.0.0.1"),portnumber(10000)).connect(SigC::slot(*this,&testbed::connected));
  };
  void dod()
  {
  };
  void connected(client_t &, client_t::result &r)
  {
    cout << "testbed connected()\n";
    if (r.isconnected()) {
      cout << "isconnected!!\n";
      con = r.getconnection();
      cout << "IP " << con->getforeign_ipv4address().dump() << "  PORT: " << con->getforeign_portnumber().getnumber() << "\n";
      con->receive.connect(SigC::slot(*this,&testbed::myrecv));
      con->sig_disconnected.connect(SigC::slot(*this,&testbed::mydeath));
    }else{
      cout << "no connection....\n";
      switch(r.geterror()) {
      case client_t::result::UNREACHABLE: cout << "UNREACHABLE\n"; break;
      case client_t::result::REFUSED: cout << "connection was refused by da host!\n"; break;
      };
    };
  };
  void myrecv(con_t &)
  {
    for (bb::index i=0;i<con->receivebuffer.size();++i) {
      cout << "[" << int(con->receivebuffer[i]) << "] ";
      if (con->receivebuffer[i]=='q') scheduler::gsched->setterminated(true);
      if (con->receivebuffer[i]=='s') {
	bb::index oldsize=con->sendbuffer.size();
	con->sendbuffer.resize(oldsize+5);
	for (bb::index i=0;i<5;++i) con->sendbuffer[oldsize+i]=i+'1';
	con->send();
      };
    };
    cout << "\n";
    con->receivebuffer.erase(0,con->receivebuffer.size());
  };
  void mydeath(con_t &)
  {
    cout << "mydeath\n";
  };
  client_t cli;
  sptr_con_t con;
};



class tc {
public:
  static bool debugtc;
  tc () { if (debugtc) {cout << "TC: default cons\n";}; };
  ~tc () { if (debugtc) {cout << "TC: destructor\n";}; };
  tc (const tc &) { if (debugtc) {cout << "TC: copy cons\n";}; };
  tc & operator= (const tc &) { if (debugtc) {cout << "TC: operator=\n";}; return *this; };
};

bool tc::debugtc=true;


class testbed2 : public SigC::Object {
public:
  testbed2()
  {
    //tc::debugtc=false;
    def.setreltime(seconds(2));
    def.send.connect(SigC::slot(*this,&testbed2::recfi));
  };
  void dod()
  {
    tc i;
    def.receive(0,i);
    //    def.setreltime(seconds(3));
    // def.receive(10,tc());
  };
  void recfi (const int &i, tc&)
  {
    cout << "recfi: int " << i << "\n";
    if (i==0) {
      scheduler::gsched->setterminated(true);
    }else{
      def.setreltime(seconds(i));
      //      def.receive(i-1,tc());
    };
  };
  sigdeferer2<const int&, tc&,scheduler> def;
};


class testbed3 : public SigC::Object {
public:
  testbed3()
  {
    sig.connect(slot(*this,&testbed3::slotter));
    //    sig.emit(static_cast<const tc>(tc()));
  };
  SigC::Signal1<void,tc&> sig;
  void dod()
  {
  };
  void slotter(tc&)
  {
  };
};

int main()
{
  memchk::resetsituation();
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      testbed t;
      t.dod();
      s.run();
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
