#include "memchk.hpp"
#include "yyypostman.hpp"
#include "yyynetwork.hpp"
#include "fatal.hpp"
#include "yyyscheduler.hpp"
#include "yyyexceptions.hpp"
#include "yyybinbuff.hpp"
#include "yyysendables.hpp"

#include <unistd.h>

using namespace yyybinbuff;
using namespace yyyscheduler;
using namespace yyypostman;
using namespace yyynetwork;
using namespace yyysmartptr;

class serverprocess : public SigC::Object
{
public:
  typedef connection<scheduler,binbuff> con_t;
  typedef smartptr<con_t> sptr_con_t;
  typedef server<scheduler,binbuff> server_t;
  typedef postman<const uint8,con_t> postman_t;

  server_t myserver;
  sptr_con_t mycon;
  postman_t *post;
  scheduler::ionotify io;

  serverprocess ()
    : post(0)
  {
    myserver.resultavailable.connect(SigC::slot(*this,&serverprocess::newconnection));
    myserver.setport(10000);
    io.fire.connect(SigC::slot(*this,&serverprocess::keypressed));
    io.setfilehandle(0,scheduler::ionotify::READ);
  };
  ~serverprocess ()
  {
    delete post; // <-- important that postman gets killed BEFORE the connection object
  };

  void newconnection(server_t &, sptr_con_t &c) {
    if (mycon.isnull()) {
      mycon=c;
      mycon->sig_disconnected.connect(SigC::slot(*this,&serverprocess::disconnected));
      post=new postman_t(&(*mycon));
      post->corruptreceived.connect(SigC::slot(*this,&serverprocess::corruptreceived));
      post->receiver1_open<int> ('x').connect(SigC::slot(*this,&serverprocess::recvint));
      post->sender1_open<int>('C');
      post->sender1_open<uint8> ('a');
      //post->sender1_open<const uint8> ('b');
      post->sender1_open<uint8&> ('c');
      post->sender1_open<const uint8&> ('d');
    }else{
      warning("i already have a connection, discarding the new one.");
    };
  };

  void disconnected(con_t &)
  {
    cout << "connection disconnected !\n";
    delete post; post=0;
    mycon.kill();
  };

  void corruptreceived(postman_t &)
  {
    cout << "postman corruptreceived ...\n";
    delete post; post=0;
    mycon.kill();
  };

  void recvint(int i)
  {
    cout << "RECVINT " << i << " !!!\n";
  };

  void keypressed(scheduler::ionotify &)
  {
    char c;
    read(0,&c,1);
    if (c=='q') {
      scheduler::gsched->setterminated(true);
    } else if (c=='w') {
      if (!post) { cout << "no connection, no send!\n"; return; };
      post->sender1<uint8>('a') ('A');
      //post->sender1<const uint8>('b') ('B');
      uint8 v('C');
      post->sender1<uint8 &>('c') (v);
      post->sender1<const uint8 &>('d') ('D');
    }else{
      if (!post) { cout << "no connection, no send!\n"; return; };
      post->sender1<int>('C') (c);
    };
  };
};

int main()
{
  memchk::resetsituation();
  {
    yyyexceptions::except_init();
    try {
      scheduler s;
      scheduler::gsched = &s;
      serverprocess serv;
      s.run();
    } catch (yyyexceptions::except_base &x) {
      fatal(x.humandump());
    } catch (const char *const x) {
      fatal(x);
    };
  };
  if (!memchk::everythingzero()) memchk::printstatistics(cout);
  
  return 0;
};
