/*

  yyybinbuff - a implementation of the binbuffer concept
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYBINBUFF_HPP
#define __YYYBINBUFF_HPP

#include "binarydefs.hpp"
#include "yyyexceptions.hpp"

#include <vector>

namespace yyybinbuff {

  // binbuffer is a model of the binbuffer concept
  class binbuff {
  public:
    except_mk(rangeerror, "value is out of the allowed range");

    typedef int32 index;
    binbuff ();
    ~binbuff ();
    inline const uint8 & operator [] (const index x) const { return buf[offs+x]; };
    inline uint8 & operator [] (const index x) { return buf[offs+x]; };
    inline const index size() const { return len; };

    // compatibility with old binbuff   FIXME remove this
    void resize (const index s);
    void erase (const index first, const index last);
    // end of compatibility section

    void add_end (const index amount);
    void remove_end (const index amount);
    void remove_begin (const index amount);
  private:
    uint8 *buf;
    index offs, len, cap;
    void collapse (index start, index length);
  };

};

#endif
