/*

  yyyexceptions - some tools for common exceptions
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYEXCEPTIONS_HPP
#define __YYYEXCEPTIONS_HPP

#include <string>

namespace yyyexceptions {

  void except_init ();

  class except_base {
  public:
    except_base ();
    except_base (const except_base &);
    except_base & operator= (const except_base &);
    virtual ~except_base();
    virtual string humandump() const = 0;
  };

  class except_normal : public except_base {
  public:
    except_normal (const string &errormessage); // need not terminating newline
    string humandump() const;
  private:
    string errormessage;
  };

  class except_codepos_internal : public except_normal {
  public:
    except_codepos_internal (const string &errormessage, const string &filename, const int linenumber, const bool disperrno=false);
  };

  // yes this probably ignores the namespace, but there seems to be no better way.
#define except_codepos(x) yyyexceptions::except_codepos_internal (x,__FILE__,__LINE__,false)
#define except_codepos_errno(x) yyyexceptions::except_codepos_internal (x,__FILE__,__LINE__,true)

#define except_mk(x,y) class x : public yyyexceptions::except_codepos_internal \
  { public: x () : except_codepos_internal(y,__FILE__,__LINE__,false) {}; };
};

#endif
